table.svelte-dsaf7t.svelte-dsaf7t{width:100%}.isSortable.svelte-dsaf7t.svelte-dsaf7t,.isClickable.svelte-dsaf7t.svelte-dsaf7t{cursor:pointer}tr.svelte-dsaf7t th select.svelte-dsaf7t{width:100%}td{height:1px}table,th,td{border-bottom:1px solid #e6e6e6;border-collapse:collapse}thead th{position:sticky;top:0;z-index:1;background:#fff;padding-block:10px;padding-inline:10px}.score-col{font-size:14px;text-align:center;height:100%}.overview-card{padding-block:10px;padding-inline:20px;border-radius:10px;border:1px solid #e6e6e6;font-size:16px;overflow-y:scroll}.overview-card-left{float:left;width:60%;height:100%}.overview-card-right{float:right;width:30%;height:100%}.highlight-card{padding-inline:20px;border-radius:10px;border:1px solid #e6e6e6;overflow-y:scroll;max-height:600px}.row{clear:both;margin-bottom:21px}.matrix-widget{display:flex;flex-direction:row;justify-content:space-between}.matrix-view{min-height:750px}.matrix{float:left;max-width:40%;overflow-x:scroll}.tables{float:right;padding:0 20px;overflow-x:scroll;max-width:60%}.tables p{font-size:14px}h2{font-size:16px}h3{font-size:14px}
