.concept-card{padding:10px;margin:5px 0;border-radius:10px;border:1px solid #e6e6e6;width:95%;max-height:25px;overflow-y:scroll}.concept-detail{display:flex;flex-direction:row;justify-content:space-between}.left.svelte-ezd55v.svelte-ezd55v{width:25%;display:inline-block}.left.svelte-ezd55v input.svelte-ezd55v{float:left}.left.svelte-ezd55v label.svelte-ezd55v{margin-left:5px}.mid.svelte-ezd55v.svelte-ezd55v{width:40%}.right.svelte-ezd55v.svelte-ezd55v{float:right;width:50%;padding-left:10px}.header.svelte-ezd55v.svelte-ezd55v{font-size:12px;font-weight:700;padding:0;margin:0}.examples.svelte-ezd55v.svelte-ezd55v{font-size:10px}.concept-detail.svelte-ezd55v p.svelte-ezd55v,ul.svelte-ezd55v.svelte-ezd55v{padding:0 10px;margin:0}
